use std::collections::HashMap;

use serde::{Deserialize, Serialize};

use crate::{Either, OneOrMany, identifier::Identifier, particle::Particle, tag::Tag};

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct Biome {
    pub has_precipitation: bool,
    pub temperature: f32,
    #[serde(default)]
    pub temperature_modifier: BiomeTemperatureModifier,
    pub downfall: f32,
    pub effects: BiomeEffects,
    pub carvers: OneOrMany<Either<Identifier, Tag>>,
    pub features: Vec<Vec<Either<Identifier, Tag>>>,
    // TODO: Limit between 0.0 and 0.9999999
    pub creature_spawn_probability: Option<f32>,
    pub spawners: BiomeSpawners,
    pub spawn_costs: HashMap<Identifier, BiomeSpawnCost>,
}

#[derive(Debug, Serialize, Deserialize, Clone, Default)]
#[serde(rename_all = "snake_case")]
pub enum BiomeTemperatureModifier {
    #[default]
    None,
    Frozen,
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct BiomeEffects {
    pub fog_color: i32,
    pub sky_color: i32,
    pub water_color: i32,
    pub water_fog_color: i32,
    pub foliage_color: Option<i32>,
    pub grass_color: Option<i32>,
    #[serde(default)]
    pub grass_color_modifier: BiomeGrassColorModifier,
    pub particle: Option<BiomeParticle>,
    pub ambient_sound: Option<Either<Identifier, SoundEvent>>,
    pub mood_sound: Option<BiomeMoodSound>,
    pub additions_sound: Option<BiomeAdditionsSound>,
    pub music: Option<Vec<BiomeMusicEntry>>,
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct SoundEvent {
    pub sound_id: Identifier,
    pub range: Option<f32>,
}

#[derive(Debug, Serialize, Deserialize, Clone, Default)]
#[serde(rename_all = "snake_case")]
pub enum BiomeGrassColorModifier {
    #[default]
    None,
    DarkForest,
    Swamp,
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct BiomeParticle {
    pub probability: f64,
    pub options: Particle,
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct BiomeMoodSound {
    pub sound: Either<Identifier, SoundEvent>,
    pub tick_delay: i32,
    pub block_search_extent: i32,
    pub offset: f64,
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct BiomeAdditionsSound {
    pub sound: Either<Identifier, SoundEvent>,
    pub tick_chance: f64,
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct BiomeMusicEntry {
    pub data: BiomeMusicData,
    pub weight: i32,
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct BiomeMusicData {
    pub sound: Either<Identifier, SoundEvent>,
    pub min_delay: i32,
    pub max_delay: i32,
    pub replace_current_music: bool,
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct BiomeSpawners {
    pub monster: Option<Vec<BiomeSpawner>>,
    pub creature: Option<Vec<BiomeSpawner>>,
    pub ambient: Option<Vec<BiomeSpawner>>,
    pub water_creature: Option<Vec<BiomeSpawner>>,
    pub underground_water_creature: Option<Vec<BiomeSpawner>>,
    pub water_ambient: Option<Vec<BiomeSpawner>>,
    pub misc: Option<Vec<BiomeSpawner>>,
    pub axolotls: Option<Vec<BiomeSpawner>>,
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct BiomeSpawner {
    #[serde(rename = "type")]
    pub ty: Identifier,
    pub weight: i32,
    #[serde(rename = "minCount")]
    pub min_count: i32,
    #[serde(rename = "maxCount")]
    pub max_count: i32,
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct BiomeSpawnCost {
    pub energy_budget: f64,
    pub charge: f64,
}
